/*
 * File:  main.cc
 * Author:  mikolas
 * Created on:  Tue, Jul 29, 2014 4:01:17 PM
 * Copyright (C) 2014, Mikolas Janota
 */
#include "auxiliary.hh"
#include "qcir_parse.tab.h"
#include "DataStructures.hh"
#include "qtypes.hh"
#include "AReQS2.hh"
#include "Options.hh"
#include <string>
using namespace std;
using namespace areqs2;
void read_qcir(string filename);
extern int qcir_line;
extern Expressions f;
extern QFla qcir_qfla;
extern areqs2::VariableManager qcir_vmng;
extern unordered_map<const char*,int,cstrHash,cstrEq> name2var;

unordered_map<int,const char*> var2name;
ostream& print_usage(const Options& options,ostream& o);

ostream& print_model(ostream& o, const SATSPC::vec<lbool>& m, const VarVector& vs) {
  FOR_EACH(i,vs){
    const auto v=*i;
    const auto val=v<m.size()?m[v]:l_Undef;
    if (val==l_Undef)continue;
    o<<' '<<(val==l_False?'-':'+')<<var2name[v];
  }
  return o;
}

int main(int argc, char** argv) {
  Options options;
  if (!options.parse(argc, argv)) {
     cerr << "ERROR: processing options." << endl;
     print_usage(options,cerr);
     return 100;
  }
  auto& rest = options.get_rest();
  if (rest.size()>1)
    cerr<<"WARNING: garbage at the end of command line."<<endl;

  if (options.get_help()){
     print_usage(options,cout);
     return 0;
  }

  const string filename(rest.size() ? rest[0] : "-");
  read_qcir(filename);
  //ExpressionPrinter p(cerr,f);
  //cerr<<qcir_qfla.pref<<endl;
  //p.visit(qcir_qfla.output);
  //cerr<<std::endl;
  auto& pref=qcir_qfla.pref;
  QFla solved_fla;
  solved_fla.pref.push_back(Quantification(EXISTENTIAL,VarVector()));
  solved_fla.pref.push_back(Quantification(UNIVERSAL,VarVector()));
  bool swap=pref.size()&&pref[0].first==UNIVERSAL;
  size_t cq=0;
  while(cq<pref.size()&&pref[cq].first==(swap ? UNIVERSAL : EXISTENTIAL)) {
    solved_fla.pref[0].second=VarVector::conc(solved_fla.pref[0].second,pref[cq].second);
    ++cq;
  }
  while(cq<pref.size()&&pref[cq].first==(swap ? EXISTENTIAL : UNIVERSAL)) {
    solved_fla.pref[1].second=VarVector::conc(solved_fla.pref[1].second,pref[cq].second);
    ++cq;
  }
  if(cq<pref.size()) {
    cerr<<"unsupported prefix, only prefix of at most 2 are supported"<<endl;
    exit(100);
  }
  FOR_EACH(i,name2var)var2name[i->second]=i->first;
  solved_fla.output=swap ? f.make_not(qcir_qfla.output) : qcir_qfla.output;
  AReQS2 a2(options,f,qcir_vmng,solved_fla,var2name);
  bool r=a2.solve();
  cerr<<"Total time "<<read_cpu_time()<<std::endl;
  if(r)print_model(cout<<"v", a2.get_model(), solved_fla.pref[0].second)<<std::endl;
  if(swap) r=!r;
  std::cout<<"s cnf "<<(r?'1':'0')<<std::endl;
  return r ? 10 : 20;
}

ostream& print_usage(const Options& options,ostream& o) {
  o << "USAGE: [OPTIONS] [FILENAME]" << endl;
  return options.print(o);
}
