/*
 * File:   minisat_aux.hh
 * Author: mikolas
 *
 * Created on October 21, 2010, 2:10 PM
 */

#ifndef MINISAT_AUX_HH
#define	MINISAT_AUX_HH
#include "auxiliary.hh"
#include "core/Solver.h"
#include <ostream>
#include "mtl/Vec.h"
namespace areqs2 {
using std::ostream;
using std::vector;
using SATSPC::lbool;
using SATSPC::mkLit;
using SATSPC::sign;
using SATSPC::var;
using SATSPC::vec;
using SATSPC::Lit;
using SATSPC::Var;
typedef std::vector<Lit>                      LiteralVector;

ostream& print_model(ostream& out, const vec<lbool>& lv);
ostream& print_model(ostream& out, const vec<lbool>& lv, int l, int r);
ostream& print(ostream& out, const vec<Lit>& lv);
ostream& printlv(ostream& out, const vec<Lit>& lv);
ostream& print(ostream& out, const vector<Lit>& lv);
ostream& print(ostream& out, Lit l);
ostream& print(ostream& out, lbool lb);
ostream& operator << (ostream& outs, Lit lit);
ostream& operator << (ostream& outs, lbool lb);
ostream& operator << (ostream& outs, LiteralVector ls);

inline ostream& operator << (ostream& outs, const vec<Lit>& lv) {
  return print(outs, lv);
}

inline Lit to_lit(Var v, lbool val) {return val==l_True ? mkLit(v) : ~mkLit(v);}

inline Lit to_lit(const vec<lbool>& bv, Var v) {
  return (v<bv.size()) && (bv[v]==l_True) ? mkLit(v) : ~mkLit(v);
}

inline lbool eval(Var v,const vec<lbool>& vals) {
  if(v>=vals.size()) return l_Undef;
  return vals[v];
}

inline lbool eval(Lit l,const vec<lbool>& vals) {
  const Var v=var(l);
  assert(v<vals.size());
  const auto vv=vals[v];
  if(vv==l_Undef)return l_Undef;
  return (vv==l_False)==(sign(l)) ? l_True : l_False;
}

inline void to_lits(const vec<lbool>& bv, vec<Lit>& output, int s, const int e) {
  for (int index = s; index <= e; ++index) {
    if (bv[index]==l_True) output.push(mkLit((Var)index));
    else if (bv[index]==l_False) output.push(~mkLit((Var)index));
  }
}

class Lit_equal {
public:
  inline bool operator () (const Lit& l1,const Lit& l2) const { return l1==l2; }
};

class Lit_hash {
public:
  inline size_t operator () (const Lit& l) const { return SATSPC::toInt(l); }
};


inline size_t literal_index(Lit l) { 
  assert(var(l) > 0);
  const size_t v = (size_t) var(l);
  return sign(l) ? v<<1 : ((v<<1)+1);
}

inline Lit index2literal(size_t l) { 
  const bool positive = (l&1);
  const Var  variable = l>>1;
  return positive ? mkLit(variable) : ~mkLit(variable);
}
};
#endif	/* MINISAT_AUX_HH */
